// CodeMirror, copyright (c) by Marijn Haverbeke and others
// Distributed under an MIT license: https://codemirror.net/5/LICENSE
(function(mod) {
  if (typeof exports == "object" && typeof module == "object") // CommonJS
    mod(require("../../lib/codemirror"));
  else if (typeof define == "function" && define.amd) // AMD
    define(["../../lib/codemirror"], mod);
  else // Plain browser env
    mod(CodeMirror);
})(function(CodeMirror) {
  'use strict';

  CodeMirror.defineMode('puppet', function () {
    let words = {};
    let variable_regex = /({)?(\w+)?((::\w+)+)?(\w+)(})?/;

    function define(style, string) {
      let split = string.split(' ');
      for (let word of split) {
        words[word] = style;
      }
    }

    define('settingsConfig', 'input filter output');
    define('keyword', 'class define site node include import inherits');
    define('keyword', 'case if else in and elsif default or');
    define('atom', 'false true running present absent file directory undef');
    define(
      'builtin',
      'action augeas burst chain computer cron destination dport exec ' +
      'file filebucket group host icmp iniface interface jump k5login limit log_level ' +
      'log_prefix macauthorization mailalias maillist mcx mount nagios_command ' +
      'nagios_contact nagios_contactgroup nagios_host nagios_hostdependency ' +
      'nagios_hostescalation nagios_hostextinfo nagios_hostgroup nagios_service ' +
      'nagios_servicedependency nagios_serviceescalation nagios_serviceextinfo ' +
      'nagios_servicegroup nagios_timeperiod name notify outiface package proto reject ' +
      'resources router schedule scheduled_task selboolean selmodule service source ' +
      'sport ssh_authorized_key sshkey stage state table tidy todest toports tosource ' +
      'user vlan yumrepo zfs zone zpool',
    );

    function tokenString(stream, state) {
      let current, prev, found_var = false;
      while (!stream.eol() && (current = stream.next()) != state.pending) {
        if (current === '$' && prev != '\\' && state.pending == '"') {
          found_var = true;
          break;
        }
        prev = current;
      }
      if (found_var) {
        stream.backUp(1);
      }
      if (current == state.pending) {
        state.continueString = false;
      } else {
        state.continueString = true;
      }
      return 'string';
    }

    function handleVariable(stream, state) {
      if (stream.match(variable_regex)) {
        return state.continueString ? 'variable-2' : 'variable';
      }
      return 'error';
    }

    function handleDefinition(stream, state) {
      if (stream.match(/(\s+)?[\w:]+(\s+)?/)) {
        return 'def';
      }
      stream.match(/\s+{/);
      state.inDefinition = false;
    }

    function handleInclude(stream, state) {
      stream.match(/(\s+)?\S+(\s+)?/);
      state.inInclude = false;
      return 'def';
    }

    function handleFunction(stream) {
      stream.backUp(1);
      return 'def';
    }

    function handleAttribute(stream) {
      stream.match(/(\s+)?\w+/);
      return 'tag';
    }

    function handleWord(stream, state, word) {
      stream.backUp(1);
      stream.match(/\w+/);
      if (stream.match(/\s+\S+\s+{/, false)) {
        state.inDefinition = true;
      }
      if (word == 'include') {
        state.inInclude = true;
      }
      return words[word];
    }

    function handleReference(stream, word) {
      stream.backUp(1);
      stream.match(/(^|\s+)[A-Z][\w:]+/);
      return 'def';
    }

    function handleResource(stream) {
      stream.match(/(\s+)?[\w:]+/);
      return 'def';
    }

    function handleSpecialResource(stream) {
      stream.match(/(\s+)?@{1,2}/);
      return 'special';
    }

    function handleComment(stream) {
      stream.skipToEnd();
      return 'comment';
    }

    function handleEscape(stream) {
      stream.next();
      stream.skipToEnd();
      return 'string';
    }

    function handleString(stream, state, ch) {
      state.pending = ch;
      return tokenString(stream, state);
    }

    function handleBracket() {
      return 'bracket';
    }

    function handleRegex(stream) {
      stream.match(/^[^\/]*\//);
      return 'variable-3';
    }

    function handleNumber(stream) {
      stream.eatWhile(/\d+/);
      return 'number';
    }

    function handleOperator(stream) {
      if (stream.peek() == '>') {
        stream.next();
      }
      return 'operator';
    }

    function tokenize(stream, state) {
      let word = stream.match(/\w+/, false);
      let attribute = stream.match(/(\s+)?\w+\s+=>.*/, false);
      let resource = stream.match(/(\s+)?[\w:]+(\s+)?{/, false);
      let special_resource = stream.match(/(\s+)?@{1,2}[\w:]+(\s+)?{/, false);
      let ch = stream.next();

      if (ch === '$') {
        return handleVariable(stream, state);
      }
      if (state.continueString) {
        stream.backUp(1);
        return tokenString(stream, state);
      }
      if (state.inDefinition) {
        return handleDefinition(stream, state);
      }
      if (state.inInclude) {
        return handleInclude(stream, state);
      }
      if (stream.match(/(\s+)?\w+\(/)) {
        return handleFunction(stream);
      }
      if (attribute) {
        return handleAttribute(stream);
      }
      if (word && words.hasOwnProperty(word)) {
        return handleWord(stream, state, word);
      }
      if (/(^|\s+)[A-Z][\w:]+/.test(word)) {
        return handleReference(stream, word);
      }
      if (resource) {
        return handleResource(stream);
      }
      if (special_resource) {
        return handleSpecialResource(stream);
      }
      if (ch == '#') {
        return handleComment(stream);
      }
      if (ch == '\\') {
        return handleEscape(stream);
      }
      if (ch == "'" || ch == '"') {
        return handleString(stream, state, ch);
      }
      if (ch == '{' || ch == '}') {
        return handleBracket();
      }
      if (ch == '/') {
        return handleRegex(stream);
      }
      if (ch.match(/\d/)) {
        return handleNumber(stream);
      }
      if (ch == '=') {
        return handleOperator(stream);
      }
      stream.eatWhile(/[\w-]/);
      return null;
    }

    return {
      startState: function () {
        let state = {};
        state.inDefinition = false;
        state.inInclude = false;
        state.continueString = false;
        state.pending = false;
        return state;
      },
      token: function (stream, state) {
        if (stream.eatSpace()) return null;
        return tokenize(stream, state);
      },
    };
  });

  CodeMirror.defineMIME('text/x-puppet', 'puppet')
})
